package BNF

/*
Syntax of the output operator trees:

CLIST ::=  List( CTREE+ )
             that is, a list of zero or more  CTREEs
DLIST ::=  List( DLIST+ )

DTREE ::=  Var( Iden, ETREE )
        |  Proc( Iden, ILIST, CLIST, Option( ETREE ) )
        |  Class( Iden, TTREE )

CTREE ::=  Dec( DTREE )  |  Assign( LTREE, ETREE )  
        |  CallProc( LTREE, ELIST )
        |  If( ETREE, CLIST, CLIST )  |  While( ETREE, CLIST )
        |  Print( ETREE )

ETREE ::=  Num( N )  |  Binop( OP, ETREE, ETREE )  |  Deref( LTREE )
        |  New( TTREE )  |  CallFunc( LTREE, ELIST )

LTREE ::=  This()  |  Super()  |  Ref( Iden ) |  Dot( LTREE, Iden )

TTREE ::=  CallClass( LTREE )  |  Template( Option(Iden), DLIST, CLIST )

OP ::=  "+"  |  "-"  | ...        
ILIST ::=  List( I* )   that is, a list of zero or more  Is
ELIST ::=  List( ETREE* )
*/

sealed abstract class Dtree
  case class Var(name: String, e: Etree) extends Dtree
  case class Proc(name: String, params: List[String], 
                  body: List[Ctree], result: Option[Etree] ) extends Dtree
  case class Class(name: String, body: Ttree) extends Dtree
  
sealed abstract class Ctree
  case class Dec(d: Dtree) extends Ctree
  case class Assign(l: Ltree, e: Etree) extends Ctree
  case class If(e: Etree, th: List[Ctree], el: List[Ctree]) extends Ctree
  case class While(e: Etree, c: List[Ctree]) extends Ctree
  case class Print(e: Etree) extends Ctree
  case class CallProc(f: Ltree, args: List[Etree]) extends Ctree
 
sealed abstract class Etree
  case class Num(n: String) extends Etree
  case class BinOp(op: String, e1: Etree, e2: Etree) extends Etree
  case class Deref(l: Ltree) extends Etree
  case class New(t: Ttree) extends Etree
  case class CallFunc(f: Ltree, args: List[Etree]) extends Etree
  
sealed abstract class Ltree
  case class This() extends Ltree
  case class Super() extends Ltree
  case class Ref(id: String) extends Ltree
  case class Dot(objectPart: Ltree, fieldPart: String) extends Ltree

sealed abstract class Ttree
  case class Template(superclass: Option[String], 
                      privateDecs: List[Dtree], 
                      body: List[Ctree]) extends Ttree
  case class CallClass(f: Ltree) extends Ttree
