
/*
   Interpreter for mini object language with procedures.
   The interpreter interprets this abstract syntax:
   (see  BNF.scala  for the class definitions)

CLIST ::=  List( CTREE,* )
             that is, a list of zero or more  CTREEs
DLIST ::=  List( DLIST* )

DTREE ::=  Var( Iden, ETREE )
        |  Proc( Iden, ILIST, CLIST, Option( ETREE ) )
        |  Class( Iden, TTREE )      
  Coded like this:
      sealed abstract class Dtree
        case class Var(name: String, e: Etree) extends Dtree
        case class Proc(name: String, params: List[String], 
            body: List[Ctree], result: Option[Etree] ) extends Dtree 
        case class Class(name: String, body: Ttree) extends Dtree

CTREE ::=  Dec( DTREE )  |  Assign( LTREE, ETREE )  
        |  CallProc( LTREE, ELIST )
        |  If( ETREE, CLIST, CLIST )  |  While( ETREE, CLIST )
        |  Print( ETREE )

ETREE ::=  Num( N )  |  Binop( OP, ETREE, ETREE )  |  Deref( LTREE )
        |  New( TTREE )  |  CallFunc( LTREE, ELIST )

LTREE ::=  This()  |  Super()  |  Ref( Iden ) |  Dot( LTREE, Iden )

TTREE ::=  CallClass( LTREE )  
        |  Template( Option(Iden), DLIST, CLIST )

OP ::=  "+"  |  "-"  | ...        
ILIST ::=  List( I,* )   that is, a list of zero or more  Is
ELIST ::=  List( ETREE,* )
*/

package Interpreter

import BNF._       // definitions of syntax trees in  BNF.scala
import VM.Heap._   // virtual machine: heap + activation stack
import VM.Stack._


object Interpreter {

// MAIN INTERPRETATION FUNCTION:

def interpret(p: Ttree): Unit =  {
    /* interprets a complete program tree
       param: tree is a program tree of form  PROGRAM ::=  TEMPLATE */
    println("Interpreter started:")
    printHeap()
    interpretTtree(p)
    println("Terminated.  Final configuration:")
    printHeap()
}



def interpretDtree(d: Dtree) : Unit = {
    /* param: d  is a declaration represented as a Dtree:
       DTREE ::=  Var( Iden, ETREE )
               |  Proc( Iden, ILIST, CLIST, Option( ETREE ) )
               |  Class( Iden, TTREE )
    */
    d match {
        case Var(i, e) => bind(top(), i, interpretEtree(e))
        case Proc(name, params, body, result) =>  println("WRITE Proc Dec")
        case Class(name, body) => println("WRITE Class Dec IN ASSN3")
    }
}


def interpretCtree(c: Ctree) : Unit = {
    /* param: c  is a command represented as a CTREE:
       CTREE ::=  Dec( DTREE )  |  Assign( LTREE, ETREE )  
               |  CallProc( LTREE, ELIST )
               |  If( ETREE, CLIST, CLIST )  |  While( ETREE, CLIST )
               |  Print( ETREE )
    */
    c match {
       case Dec(d) =>  interpretDtree(d)
       case Assign(l,e) => {
              val (handle, fieldname) = interpretLtree(l)
              bind(handle, fieldname, interpretEtree(e))
       }
       case While(e,cs) => {
              interpretEtree(e) match {
                  case Value(n) => {
                      if (n != 0) {
                          interpretClist(cs)
                          interpretCtree(c)  //loop more
                      }
                  }
                  case _=>  crash("non Int used in while-test")
              }
       }
       case CallProc(l, elist) => println("WRITE proc call")
       case If(e, thenlist, elselist) => println("WRITE if command")
       case Print(e) => {
             print("Print: ")
             println(interpretEtree(e))
             printHeap() 
       }
    }
}


def interpretClist(clist: List[Ctree]) : Unit = {
    /* param: clist  is a list of  CTREEs  */
    for (command <- clist) interpretCtree(command)
}



def interpretEtree(etree: Etree) : Meaning = {
    /* param: etree  is a tree of form
       ETREE ::=  Num( N )  |  Binop( OP, ETREE, ETREE )  |  Deref( LTREE )
               |  New( TTREE )  |  CallFunc( LTREE, ELIST )
       where  OP is "+" or "-"
       returns: Meaning (handle or int or Nill), the meaning of  etree
    */
    etree match {
        case Num(n) => Value(n.toInt)
        case BinOp(op, e1,e2) => {
                val n1: Int = interpretEtree(e1) match {
                             case Value(n) => n
                             case _ => { crash("arithmetic error")
                                         throw new Exception() }
                         }
                val n2: Int = interpretEtree(e2) match {
                             case Value(n) => n
                             case _ => { crash("arithmetic error")
                                         throw new Exception() }
                         }
                Value(op match {
                        case "+" => n1 + n2
                        case "-" => n1 - n2
                        case _ => { crash("arithmetic error")
                                    throw new Exception() }
                      })
        }
        case Deref(l) => {
                val (handle, fieldname) = interpretLtree(l)
                find(handle, fieldname)
        } 
        case New(t) => interpretTtree(t)
        case CallFunc(l, args) => println("WRITE function call"); Nill
    }
}



def interpretClist(elist: List[Etree]) : List[Meaning]= {
    /* param: elist  is a list of exprs,   ELIST ::=  [ ETREE* ]
       returns: a list of the values of the members of  elist */
    for (expr <- elist) yield interpretEtree(expr)  // builds list
}



def interpretTtree(t: Ttree):  Handle  = {
    /* param: t is a template of form, 
       TTREE ::=  CallClass( LTREE )
               |  Template( Option(Iden), DLIST, CLIST )
       returns: the handle to the object constructed from  t
    */
    t match {
        case Template(superclassname, privates, body) => {
                // for now, ignore first two args!  CODE THEM IN ASSN3.
                val newhandle = alloc(Map("parent"->top()))
                push(newhandle)
                interpretClist(body)
                pop()
                newhandle
        }
        case CallClass(ltree) => println("WRITE class call IN ASSN3")
                                 alloc(Map())  // REMOVE ME LATER
    }
}



def interpretLtree(lt: Ltree) : (Handle, String) = {
    /* LTREE ::=  This()  |  Super()  |  Ref( Iden ) |  Dot( LTREE, Iden )
       returns Lval meaning of lt, a  handle, fieldname  pair
    */
    lt match {
        case This() =>
              if (member("this", top())) (top(), "this")
              else  { crash("invalid use of 'this'");  (new Handle(-1), "") }
        case Super() => 
              if (member("super", top())) (top(), "super")
              else  { crash("invalid use of 'super'");  (new Handle(-1), "") }
        case Ref(id) => 
              searchStaticChain(top(), id)
        case Dot(ltree, fieldname) => {
              val (objecthandle, itsfield) = interpretLtree(ltree)
              find(objecthandle, itsfield) match {
                case Handle(h) => (Handle(h), fieldname)
                case _ =>
                   { crash("handle lookup error");  throw new Exception() }
              }
        }
    }
}


def searchStaticChain(han: Handle, i: String) : (Handle, String) = {
    /* searches for iden  i,  starting in object named by  han.
       If not found, searches parent links till  i  found (or error)
    */
    if (member(i, han))  (han, i)
    else
       find(han, "parent") match {
             case Handle(loc) => searchStaticChain(Handle(loc), i)
             case _ => { crash("iden " + i + "not found")
                         throw new Exception() }
       }
}



def crash(message: String) : Unit = {
    /* prints an error message and stops the interpreter
       params: message - a string to print
               tree - the parse tree that caused the error
    */
    println("ERROR: " + message)
    //print("At phrase:");  println(tree)
    printHeap()
    //throw new Exception
}

}

