/* Virtual machine for object language: heap and stack */

package VM

import scala.collection.mutable



object Heap { // heap module:
/* THE HEAP:

  heap : { Handle : Namespace },  i.e., a Map

   where Namespace = { Identifier : Meaning }
         Meaning = Handle | Integer | Nil (no value)
         LValue = Handle * Identifier
         Identifier = String
*/
sealed abstract class Meaning
case class Handle(loc: Int) extends Meaning
case class Value(n: Int) extends Meaning


/* the Heap maps handles to namespaces: */
var heap: mutable.Map[Handle, mutable.Map[String, Meaning]] 
        = mutable.Map()

  var heapCounter = 0    // used to generate new handles
  
  val Nill: Handle = Handle(-1)  // a hack used to generate a "nil" address
                         //  (in Scala, Nil is the empty list)

                         
  def genHandle() : Handle = {
    /* generates and returns a new handle value */ 
    val han = heapCounter
    heapCounter = heapCounter + 1
    Handle(han)
  }
  
  
  def alloc(d: Map[String, Meaning]): Handle =  {
    /* allocates a new object in the heap
         by *copying*  d's  contents into the new object
       param d - a Map holding the bindings to be copied
       returns - the handle to the new object  */ 
    val han = genHandle()  // see helper function just above
    var m : mutable.Map[String,Meaning] = mutable.Map()
    for(k <- d.keys) { m(k) = d(k) }  // copy  d's  contents into  m
    heap += (han -> m)  // adjoin new Map into heap
    return han  // return the handle of newly allocated object
  }


  def bind(handle: Handle, fieldname: String, meaning: Meaning): Handle = {
    /* binds fieldname to meaning in the namespace labelled by handle
       returns  handle,  if successful  */
    if (heap contains handle) { 
      heap(handle)(fieldname) = meaning
      return handle
    }
    else // the handle is not in the heap:
      throw new Exception("bind error: " + handle + " does not exist in the heap")
  }

  
  def find(handle: Handle, fieldname: String): Meaning = {
    /* finds and returns the value bound to fieldamane in the namespace 
       labelled by handle */
    if (heap contains handle) 
      return heap(handle)(fieldname)
    else 
      throw new Exception("find error: " + handle + " does not exist in the heap")
  }

  
  def member(fieldname: String, handle: Handle): Boolean = {
    /* returns whether name is bound in the namespace labelled by handle*/
    (heap contains handle) && (heap(handle) contains fieldname)
  }  
  
  // see end of file for  printHeap  method!
  
} // end object Heap



object Stack {  // stack module:
/* ACTIVATION STACK:  uses the heap
    modelled as a linked list of namespaces of form,

    {'ns' -> Handle, 'parent' -> Handle}
      where var  actstack  points to the front namespace of the list
*/
import Heap._

var actstack: Handle = Nill  // stack starts empty


  def push(h: Handle): Unit = {
    /* makes the namespace labelled  h  the active namespace.
       param: h - the handle to be pushed onto the stack */
    actstack = alloc(Map("ns"-> h, "parent"-> actstack))
  } 

  def pop(): Unit = {
    /* forgets the active namespace by popping the activation stack */
    actstack = 
       find(actstack, "parent").asInstanceOf[Handle]
  } 

  def top(): Handle = {
    /* returns the handle of the active namespace from the stack top */
    if (actstack == Nill )  // stack empty ?
        return Nill  // so it goes...
    else  return find(actstack, "ns").asInstanceOf[Handle]
  }

  
  def printHeap() : Unit = {
    /* prints the contents of the heap, slightly formatted */
    
    def printMeaning(m: Meaning) : Unit = {
        m match {
           case Handle(h) => 
              if (m eq Nill)  print("Nil")
              else  print("H" + h) 
           case Value(n) => print(n)
        }
    }
    def printNS(n: mutable.Map[String, Meaning]) : Unit = {
        print("Map(")
        val hlist = n.keys
        for (k <- hlist) { print(k + "-> ")
                           printMeaning(n(k))
                           print(", ") }
        print(")") 
    }
    print("actstack = ");  printMeaning(actstack);  println()
    println("heap = Map(")
    val hlist = heap.keys
    //hlist.sort()   // not available in Scala )-:
    for (k <- hlist) { 
       print("  ");  printMeaning(k);  print("-> ");
       printNS(heap(k));  println()}
    println(")")
  }
  
} // end object Stack
