
# PLY template of parser for baby object language.

"""
Here is the syntax that is parsed:

CL: CommandList
D : Declaration
C : Command                 L : LefthandSide
E : Expression              T : Template
I : Variable                N : Numeral
IL: VariableList            EL: ExpressionList

CL ::= sequence of zero or more  C,  separated by semicolons

D ::=  var I = E 
    |  proc I ( IL ) : CL /? return E ?/ end 
    |  class I : T
    where  /? ... ?/  means that ... is optional
DL ::= sequence of zero or more  D,  separated by semicolons 

C ::=  D  |  L = E  |  L ( EL )  |  if E : CL else CL end 
    |  while E : CL end  |  print E

E ::=  N  |  ( E1 OP E2 )  |  L  |  L ( EL )  |  new T 
       where  OP  is  '+'  or  '-'

L ::=  this  |  super |  I  |  L . I

T ::=   L  |  /? extends I with ?/ /? private DL ?/ { CL } 

IL ::= sequence of zero or more  I,  separated by commas
EL ::= sequence of zero or more  E,  separated by commas
N ::=  string of digits
I ::=  string of letters, not including keywords ('var', 'if', ..., 'this', ...)


Syntax of the output operator trees:

CLIST ::=  List( CTREE,* )
             that is, a list of zero or more  CTREEs
DLIST ::=  List( DLIST* )

DTREE ::=  Var( Iden, ETREE )
        |  Proc( Iden, ILIST, CLIST, Option( ETREE ) )
        |  Class( Iden, TTREE )      
    that is, generated from
      sealed class DTREE
        case class Var(name: String, e: ETREE) extends DTREE
        case class Proc(name: String, params: List[String], 
            body: List[CTREE], result: Option[ETREE] ) extends DTREE 
        case class Class(name: String, body: TTREE) extends DTREE

CTREE ::=  Dec( DTREE )  |  Assign( LTREE, ETREE )  
        |  CallProc( LTREE, ELIST )
        |  If( ETREE, CLIST, CLIST )  |  While( ETREE, CLIST )
        |  Print( ETREE )

ETREE ::=  Num( N )  |  Binop( OP, ETREE, ETREE )  |  Deref( LTREE )
        |  New( TTREE )  |  CallFunc( LTREE, ELIST )

LTREE ::=  This()  |  Super()  |  Ref( Iden ) |  Dot( LTREE, Iden )

TTREE ::=  CallClass( LTREE )  
        |  Template( Option(Iden), DLIST, CLIST )

OP ::=  "+"  |  "-"  | ...        
ILIST ::=  List( I,* )   that is, a list of zero or more  Is
ELIST ::=  List( ETREE,* )
"""

from ply import *
import ex23lex

tokens = ex23lex.tokens

# set precedences of infix tokens like this:
#precedence = ( #('left', 'TIMES','DIVIDE'),  etc.

######################################################
# Grammar rules and their corresponding operator trees

### THE FIRST GRAMMAR RULE DEFINES THE START (MAIN) NONTERMINAL:
# P ::= T
# PTREE ::= TTREE
def p_Program(p):
    '''Program : Template''' 
    p[0] = p[1]

# CL ::= C;*  (that is,   CL ::=  C | C ; CL | empty )
# CLIST ::= List( CTREE )
def p_CommandList(p):
    '''CommandList : Command
                   | Command ';' CommandList
                   | empty '''   
    if len(p) == 2 and p[1] != None :  # CommandList : Command
       p[0] =  p[1] + "\n :: List()"
    elif len(p) == 4 :  # CommandList :  Command ';' CommandList
       p[0] = p[1] + "\n :: " + p[3]
    else :  # CommandList : empty,  because p[1] == None
       p[0] = "List()"


# D ::=  var I = E  | proc I ( IL ) : CL /? return E ?/ end 
#     |  class I : T
#DTREE ::=  Var( Iden, ETREE )
#        |  Proc( Iden, ILIST, CLIST, Option( ETREE ) )
#        |  Class( Iden, TTREE )
def p_Declaration1(p):
    '''Declaration : var Iden '=' Expression '''
    p[0] = "Var(" + p[2] + ", " + p[4] + ")"

def p_Declaration2(p):
    '''Declaration : proc Iden '(' IdentifierList ')' ':' CommandList end '''
    p[0] = "Proc(" + p[2] + ", " + p[4] + ", " + p[7] + ", None)"

def p_Declaration3(p):
    '''Declaration : proc Iden '(' IdentifierList ')' ':' CommandList return Expression end '''
    p[0] = "Proc(" + p[2] + ", " + p[4] + ", " + p[7] + ", Some(" + p[9] + "))"

def p_Declaration4(p):
    '''Declaration : class Iden ':' Template '''
    p[0] = "Class(" + p[2] + ", " + p[4] + ")"


# DL ::= D;*  (that is,   DL ::=  D | D ; CL | empty )
# DLIST ::= List( DLIST )
def p_DeclarationList(p):
    '''DeclarationList : Declaration
                   | Declaration ';' DeclarationList
                   | empty '''    
    if len(p) == 2 and p[1] != None :  # Declarationlist :  Declaration ';' Declarationlist
       p[0] =  p[1] + "\n :: List()"
    elif len(p) == 4 :  # Declarationlist :  Declaration ';' Declarationlist
       p[0] = p[1] + "\n :: " + p[3]
    else :  # DeclarationList : empty,  because p[1] == None
       p[0] = "List()"


# T ::=   L |  /? extends I with ?/ /? private DL ?/ { CL } 
# TTREE ::=  CallClass( LTREE )  
#         |  Template( Option(Iden), DLIST, CLIST )

def p_Template1(p):
    '''Template : LefthandSide'''
    p[0] = "CallClass(" + p[1] + ")"

def p_Template2(p):
    '''Template : Extendspart Privatepart '{' CommandList '}' '''
    p[0] = "Template(" +  p[1] + ", " + p[2] + ",\n " + p[4] + ")"

def p_Extends1(e) :
    '''Extendspart : empty'''
    e[0] = "None"

def p_Extends2(e) :
    '''Extendspart : extends Iden with '''
    e[0] = "Some(" + e[2] + ")"

def p_Private1(t) :
    '''Privatepart : empty'''
    t[0] = "List()"

def p_Private2(t) :
    '''Privatepart : private DeclarationList'''
    t[0] = t[2]



# C ::=  D  |  L = E  |  if E : CL1 else CL2 end  |  while E : CL end  |  print E  |  L ( EL )
#
# CTREE ::=  Dec( DTREE )  |  Assign( LTREE, ETREE )
#       |  CallProc( LTREE, ELIST )
#       |  If( ETREE, CLIST, CLIST )  |  While( ETREE, CLIST )
#       |  Print( ETREE )

def p_Command1(c):
    '''Command : LefthandSide '=' Expression'''
    c[0] = "Assign(" + c[1] + ", " +  c[3] + ")"

def p_Command2(c):
    '''Command : if Expression ':' CommandList else CommandList end '''
    c[0] = "If(" + ", " +  c[2] + ", " + c[4] + ", " +  c[6] + ")"

def p_Command3(c):
    '''Command : print Expression'''
    c[0] = "Print(" + c[2] + ")"

def p_Command4(c):
    '''Command : LefthandSide '(' ExpressionList ')' '''
    c[0] = "CallProc(" + c[1] + ", " + c[3] + ")"

def p_Command5(c):
    '''Command : Declaration '''
    c[0] = "Dec(" + c[1] + ")"

def p_Command6(c):
    '''Command : while Expression ':' CommandList end '''
    c[0] = "While(" + c[2] + ", " + c[4] + ")"


# EL ::= E,*  (that is,   EL ::=  E ETAIL | empty   
#                         ETAIL ::=  , E ETAIL  |  empty    )
# ELIST ::= List ( ETREE )
def p_ExpressionList1(t):
    '''ExpressionList : Expression EListTail'''
    t[0] = t[1] + "\n :: " + t[2]

def p_ExpressionList2(t):
    '''ExpressionList : empty'''
    t[0] = "(List(): List[Etree])"

def p_EListTail1(t):
    '''EListTail : ',' Expression EListTail'''
    t[0] = t[2] + "\n :: " + t[3]

def p_EListTail2(t):
    '''EListTail : empty'''
    t[0] = "List()"



# IL ::= I,*  (that is,   IL ::=  I ITAIL | empty   
#                         ITAIL ::=  , I ITAIL  |  empty    )
# ILIST ::= List ( String )
def p_IdentifierList1(t):
    '''IdentifierList : Iden IListTail'''
    t[0] = t[1] + "\n :: " + t[2]

def p_IdentifierList2(t):
    '''IdentifierList :  empty'''
    t[0] = "(List(): List[String])"

def p_IListTail1(t):
    '''IListTail : ',' Iden IListTail'''
    t[0] = t[2] + "\n :: " + t[3]

def p_IListTail2(t):
    '''IListTail : empty'''
    t[0] = "List()"


# E ::= SE  |  SE OP SE   (this trick allows outermost parens to be omitted)
# SE ::=  N  |  ( SE1 OP SE2 )  |  L  |  new T  |  L ( EL )
#ETREE ::=  Num( N )  |  Binop( OP, ETREE, ETREE )  |  Deref( LTREE )
#        |  New( TTREE )  |  CallFunc( LTREE, ELIST )
# where  OP ::=  "+"  |  "-"

def p_Expression1(e):
    '''Expression : SubExpression'''
    e[0] = e[1]

def p_Expression2(e):
    ''' Expression : SubExpression Op SubExpression '''
    e[0] = "BinOp(" + e[2] + ", " + e[1] + ", " +  e[3] + ")"
    
def p_SubExpression1(e):
    '''SubExpression : NUM'''
    e[0] = 'Num("' + e[1] + '")'

def p_SubExpression2(e):
    '''SubExpression : '(' SubExpression Op SubExpression ')' '''
    e[0] =  "BinOp(" + e[3] + ", " + e[2] + ", " +  e[4] + ")"

def p_SubExpression3(e):
    '''SubExpression : LefthandSide '''
    e[0] = "Deref(" + e[1] + ")"

def p_SubExpression4(e):
    '''SubExpression : new Template '''
    e[0] = "New(" + e[2] + ")"

def p_SubExpression5(e):
    '''SubExpression : LefthandSide '(' ExpressionList ')' '''
    e[0] = "CallFunc(" + e[1] + ", " + e[3] + ")"

def p_op1(o):
    '''Op : '+' '''
    o[0] = '"+"'

def p_op2(o):
    '''Op : '-' '''
    o[0] = '"-"'


# L ::=  I  |  L . I  | this  |  super 
# LTREE ::=   This()  |  Super()  |  Ref( Iden ) |  Dot( LTREE, Iden )
def p_LefthandSide1(t):
    '''LefthandSide : Iden'''
    t[0] = "Ref(" + t[1] + ")"

def p_LefthandSide2(t):
    '''LefthandSide : LefthandSide '.' Iden'''
    t[0] = "Dot(" + t[1] + ", " + t[3] + ")"

def p_LefthandSide3(t):
    '''LefthandSide : super '''
    t[0] = "Super()"

def p_LefthandSide4(t):
    '''LefthandSide : this '''
    t[0] = "This()"

    
    
# II ::= ID  (quotes the iden)
def p_Iden(t):
    '''Iden : ID '''
    t[0] = '"' + t[1] + '"'

    
#### Generic empty construction --- no token at all
def p_empty(p):
    '''empty : '''
    pass   # retains  p[0] = None


#### Parse Error handler : prints location of error and quits parser
def p_error(p):
    print "SYNTAX ERROR at LexToken(grammar_phrase, input_word, line_number, char_number):"
    print (15 * " "), p
    print "PARSER QUITS."
    raise Exception

bparser = yacc.yacc()

def parse(data,debug=0):
    bparser.error = 0
    p = bparser.parse(data,debug=debug)
    if bparser.error: return None
    return p

